#! /usr/bin/env python
'''
    resistivity.py
    Eric Ayars
    9/4/20

    A Python solution to the previously LabVIEW-based 
    "Temperature Coefficient of Resistivity" lab.

    The way this works:
        There are functions that define what to do when controls are pressed.
        There is a bit of code that draws the controls and puts it on the 
            window in the right places.
        Then the program just sits there until something happens.
            If the start button is pressed, it does start_clicked(), for example.
            The most important part is actually the timer!
            The QTimer starts when start_clicked() happens. This sets up a thread 
            that interrupts the program every two seconds and has it getData().
            getData() reads the meters and converts the thermistor resistance to 
                temperature, etc.
'''

########################################
# 
# Students set these values
#
########################################

# constants

''' These are the coefficients for the Steinhart-Hart equations that Dr. Ayars
    put into this program:
    C1 = 9.8216e-4
    C2 = 2.0396e-4
    C3 = 1.4055e-7
    Do you trust Dr. Ayars? You've got to ask yourself one question: 'Do I feel 
    lucky?' Well, do you, punk?*

    If you're not feeling lucky, here's how you'd calibrate the thermistor you've
    been given.
    1)  Measure the resistance at three known temperatures. Two easy ones would
        be freezing and boiling water: you have ice and a stove available, right?
        Room temperature might work for a third, but use your imagination.
    2)  The Steinhart-Hart equation for thermistors goes like this: (Pardon the 
        LaTeX, but you should know LaTeX and if you don't then just look up the
        equation online.)

        \[ \frac{1}{T} = C1 + C2 \ln(R) + C3 (\ln(R))^3 \]

    3)  Using the three known temperatures and the corresponding measured resistances,
        set up a system of three linear equations and solve them for {C1, C2, C3}.

    * With apologies to Dirty Harry. I don't mean to damage his reputation by using 
    his lines in physics homework.

'''

C1 = 9.8216e-4
C2 = 2.0396e-4
C3 = 1.4055e-7

# constants
MaxN = 10000                    # max number of data points before a crash!
timeMS = 2000                   # time interval for sampling, ms

########################################
# 
# And leave the rest alone?
#
########################################

import sys                      # useful system calls

import numpy as np              # arrays, etc.

# VISA stuff for communication with DMM
import pyvisa

# GUI stuff
from PyQt5.QtWidgets import QApplication, QWidget, QPushButton, QVBoxLayout, QHBoxLayout, QFileDialog, QComboBox
from PyQt5.QtCore import QTimer 

# plotting package bits
import matplotlib as mpl
mpl.use('QT5Agg')
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure

# plotting OOP stuff stolen from 
# https://www.learnpyqt.com/courses/graphics-plotting/plotting-matplotlib/
class MPLCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super(MPLCanvas, self).__init__(fig)

# global variables
timer = QTimer()                # interval-timing thread
timeNow = 0.0                   # Keep track of time
N = 0                           # current extent of data
DMM1 = None                     # For keeping track of conversations
DMM2 = None
meterSelected = False           # for keeping track of whether it's ok to start or not

def start_clicked():
    global N
    # disable start button
    startButton.setDisabled(True)
    # enable stop button
    stopButton.setDisabled(False)
    # set data-pointer to beginning of array
    N = 0
    # Start intervalometer
    timer.start(timeMS)

def stop_clicked():
    # stop the timer
    timer.stop()
    # disable stop button
    stopButton.setDisabled(True)
    # enable start button
    startButton.setDisabled(False)
    # enable save button
    saveButton.setText("Save Data")
    saveButton.setDisabled(False)

def save_clicked():
    # open datafile
    filename, _ = QFileDialog.getSaveFileName()
    file = open(filename, 'w')
    # write file header
    file.write("# Time\tTemperature\tResistance\n")
    # write all data (up to N)
    for j in range(N):
        file.write("%0.2f\t%0.5f\t%0.5f\n" % (time[j], data0[j], data1[j]))
    # close file
    file.close()
    # update save button text and availability
    saveButton.setText("Data Saved")
    saveButton.setDisabled(True)

def selectionChange1():
    # change first interface as selected
    global DMM1, meterSelected
    DMM1 = rm.open_resource(selectMenu1.currentText())
    DMM1.write("system:beep")
    if meterSelected:
        startButton.setDisabled(False)
    else:
        meterSelected = True

def selectionChange2():
    # change first interface as selected
    global DMM2, meterSelected
    DMM2 = rm.open_resource(selectMenu2.currentText())
    DMM2.write("system:beep")
    if meterSelected:
        startButton.setDisabled(False)
    else:
        meterSelected = True

def read4WireResistanceFromDMM(DMM):
    # This reads 4-wire resistance: Make sure the DMM is connected correctly
    # for this measurement! One pair of red/black wires should be attached to
    # the 'sense' input on the front of the DMM, the other pair should be 
    # attached to the 'input' input. The second pair supplies the current, the
    # first pair then measures the voltage across the actual device.
    # See handout explaining more about 4-wire measurements.
    return float(DMM.query("measure:fresistance?"))

def readResistanceFromDMM(DMM):
    # This reads 2-wire resistance, which is probably what you're used to from
    # past experience in introducutory physics courses, and handheld DMMs.
    return float(DMM.query("measure:resistance?"))

def convertResistanceToTemperature(R):
    lr = np.log(R)
    T = 1.0/(C1+C2*lr + C3*lr**3)
    return T

def getData():
    ''' 
    Note that this function uses a crappy way of doing things: it 
    just clears the entire dataset from the graph and regraphs 
    everything each time another data point shows up. This works 
    because we're collecting data slowly in this case, but there
    needs to be a better way of doing this for faster experiments!
    '''
    global timeNow, N, scene
    # update time
    timeNow += timeMS/1000.0

    # measure a pair of data points and add data to arrays. Move pointer
    # N to the end of the actual data. This will break if more than MaxN 
    # points are collected. 

    # Here's the resistance of the Pt wire
    data1[N] = read4WireResistanceFromDMM(DMM1)
    # Here's the temperature
    data0[N] = convertResistanceToTemperature(readResistanceFromDMM(DMM2))

    # Update the time (hey, it might be useful later in the data file)
    time[N] = timeNow

    # update where the next data goes in the array
    N += 1

    # clear the graph
    scene.axes.cla()                

    # redraw the graph
    scene.axes.plot(data0[:N], data1[:N], 'r.')

    # show the graph
    scene.draw()

#
# Main program
#

data0 = np.zeros(MaxN)
data1 = np.zeros(MaxN)
time = np.zeros(MaxN)

app = QApplication(sys.argv)
widget = QWidget()

#
# Interface selections
#

selections = QHBoxLayout()

# Check available VISA resources
rm = pyvisa.ResourceManager()
devices = rm.list_resources()

# first interface selection drop-down
selectMenu1 = QComboBox(widget)
selectMenu1.addItem('4-wire Pt wire meter')
selectMenu1.addItems(devices)
selectMenu1.currentIndexChanged.connect(selectionChange1)

# second interface selection drop-down
selectMenu2 = QComboBox(widget)
selectMenu2.addItem('Thermistor meter')
selectMenu2.addItems(devices)
selectMenu2.currentIndexChanged.connect(selectionChange2)

# put drop-downs in the container
selections.addWidget(selectMenu1)
selections.addWidget(selectMenu2)
selections.addStretch(1)

#
# Control buttons
#

controls = QHBoxLayout()

# start
startButton = QPushButton(widget)
startButton.setText("Start")
startButton.setDisabled(True)
startButton.clicked.connect(start_clicked)

# stop
stopButton = QPushButton(widget)
stopButton.setText("Stop")
stopButton.clicked.connect(stop_clicked)
stopButton.setDisabled(True)

# save
saveButton = QPushButton(widget)
saveButton.setText("Save Data")
saveButton.clicked.connect(save_clicked)
saveButton.setDisabled(True)

# put buttons in the container
controls.addWidget(startButton)
controls.addWidget(stopButton)
controls.addWidget(saveButton)
controls.addStretch(1)

#
# scene (graph)
# 

scene = MPLCanvas(width=5, height=4, dpi=100)

# graph toolbar
toolbar = NavigationToolbar2QT(scene, widget)

#
# window layout
#

layout = QVBoxLayout()
layout.addLayout(selections)
layout.addLayout(controls)
layout.addWidget(toolbar)
layout.addWidget(scene)

# Draw the screen
widget.setLayout(layout)
widget.setWindowTitle("Resistance v. Temperature Experiment")
widget.show()

# What to do every time the timer goes off
timer.timeout.connect(getData)

sys.exit(app.exec_())
